import logging
import os
import pytest

from . import db
from ..fb import messenger
from ... import story, utils

logger = logging.getLogger(__name__)


def teardown_function(function):
    logger.debug('tear down!')
    story.stories_library.clear()


@pytest.fixture
def build_context():
    async def builder(mongodb, no_session=False, no_user=False):
        user = None
        if not no_user:
            user = utils.build_fake_user()
            await mongodb.set_user(user)

        if not no_session:
            session = utils.build_fake_session(user=user)
            await mongodb.set_session(session)

        story.use(mongodb)
        fb = story.use(messenger.FBInterface(token='qwerty'))

        return fb, user

    return builder


@pytest.fixture
@pytest.mark.asyncio
def open_db():
    class AsyncDBConnection:
        def __init__(self):
            self.db_interface = db.MongodbInterface(uri=os.environ.get('TEST_MONGODB_URL', 'mongo'), db_name='test')

        async def __aenter__(self):
            await self.db_interface.start()
            await self.db_interface.clear_collections()
            return self.db_interface

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            self.db_interface.stop()
            await self.db_interface.clear_collections()
            self.db_interface = None

    return AsyncDBConnection


@pytest.mark.asyncio
async def test_store_restore_session(open_db):
    async with open_db() as db_interface:
        session = utils.build_fake_session()
        session_id = await db_interface.set_session(session)
        logger.debug('session_id')
        logger.debug(session_id)
        restored_session = await db_interface.get_session(session_id=session_id)
        logger.debug('restored_session')
        logger.debug(restored_session)
        restored_session = await db_interface.get_session(user_id=session['user_id'])

        for (key, value) in restored_session.items():
            if key not in ['_id']:
                logger.debug('key {} '.format(key))
                assert session[key] == restored_session[key]


@pytest.mark.asyncio
async def test_store_restore_user(open_db):
    async with open_db() as db_interface:
        user = utils.build_fake_user()

        user_id = await db_interface.set_user(user)
        restored_user = await db_interface.get_user(id=user_id)

        assert user.items() == restored_user.items()


@pytest.mark.asyncio
async def test_create_new_session(open_db):
    async with open_db() as db_interface:
        user = await db_interface.new_user(facebook_user_id='1234567890')
        session = await db_interface.new_session(facebook_user_id='1234567890', user=user)
        assert session['facebook_user_id'] == '1234567890'
        assert 'stack' in session
        assert isinstance(session['stack'], list)
        assert session['user_id'] == user['_id']


@pytest.mark.asyncio
async def test_create_new_user(open_db):
    async with open_db() as db_interface:
        user = await db_interface.new_user(facebook_user_id='1234567890')
        assert user['facebook_user_id'] == '1234567890'

@pytest.mark.asyncio
async def test_start_should_open_connection_and_close_on_stop():
    db_interface = db.MongodbInterface(uri=os.environ.get('TEST_MONGODB_URL', 'mongo'), db_name='test')
    assert not db_interface.session_collection
    assert not db_interface.user_collection
    assert not db_interface.db

    await db_interface.start()

    assert db_interface.session_collection
    assert db_interface.user_collection
    assert db_interface.db

    await db_interface.stop()

    assert not db_interface.session_collection
    assert not db_interface.user_collection
    assert not db_interface.db



import asyncio
import logging
import pytest

from . import chat, story
from .integrations import mockdb, mockhttp
from .middlewares import location, text
from .utils import answer, build_fake_session, build_fake_user, SimpleTrigger

logger = logging.getLogger(__name__)


def teardown_function(function):
    logger.debug('tear down!')
    story.stories_library.clear()


@pytest.mark.asyncio
async def test_should_run_sequence_of_parts():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()
    user = build_fake_user()
    session = build_fake_session()

    @story.on('hi there!')
    def one_story():
        @story.part()
        def then(message):
            trigger_1.passed()

        @story.part()
        def then(message):
            trigger_2.passed()

    await answer.pure_text('hi there!', session, user)

    assert trigger_1.is_triggered
    assert trigger_2.is_triggered


@pytest.mark.asyncio
async def test_should_wait_for_answer_on_ask():
    trigger = SimpleTrigger()
    session = build_fake_session()
    user = build_fake_user()

    @story.on('hi there!')
    def one_story():
        @story.part()
        async def then(message):
            return await chat.ask('How are you?', user=message['user'])

        @story.part()
        def then(message):
            trigger.passed()

    await answer.pure_text('hi there!', session, user)

    assert not trigger.is_triggered

    await answer.pure_text('Great!', session, user)

    assert trigger.is_triggered


@pytest.mark.asyncio
async def test_should_prevent_other_story_to_start_until_we_waiting_for_answer():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()
    session = build_fake_session()
    user = build_fake_user()

    @story.on('hi there!')
    def one_story():
        @story.part()
        async def then(message):
            return await chat.ask('How are you?', user=message['user'])

        @story.part()
        def then(message):
            trigger_2.passed()

    @story.on('Great!')
    def one_story():
        @story.part()
        def then(message):
            trigger_1.passed()

    await answer.pure_text('hi there!', session, user)
    await answer.pure_text('Great!', session, user)

    assert trigger_2.is_triggered
    assert not trigger_1.is_triggered


@pytest.mark.asyncio
async def test_should_start_next_story_after_current_finished():
    trigger = SimpleTrigger()
    session = build_fake_session()
    user = build_fake_user()

    @story.on('hi there!')
    def one_story():
        @story.part()
        async def then(message):
            return await chat.ask('How are you?', user=message['user'])

        @story.part()
        def then(message):
            pass

    @story.on('Great!')
    def one_story():
        @story.part()
        def then(message):
            trigger.passed()

    await answer.pure_text('hi there!', session, user)
    await answer.pure_text('Great!', session, user)
    await answer.pure_text('Great!', session, user)

    assert trigger.is_triggered


@pytest.mark.asyncio
async def test_should_match_group_of_matchers_between_parts_of_story():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()
    session = build_fake_session()
    user = build_fake_user()

    @story.on('hi there!')
    def one_story():
        @story.part()
        def then(message):
            return [text.Any(), location.Any()]

        @story.part()
        def then(message):
            trigger_1.passed()
            return [text.Any(), location.Any()]

        @story.part()
        def then(message):
            trigger_2.passed()

    await answer.pure_text('hi there!', session, user)
    await answer.location({'lat': 1, 'lng': 1}, session, user)
    await answer.pure_text('hi there!', session, user)

    assert trigger_1.is_triggered
    assert trigger_2.is_triggered


@pytest.mark.asyncio
async def test_should_match_group_of_matchers_on_story_start():
    trigger = SimpleTrigger()
    session = build_fake_session()
    user = build_fake_user()

    @story.on(receive=['hi there!', location.Any()])
    def one_story():
        @story.part()
        def then(message):
            trigger.passed()

    await answer.pure_text('hi there!', session, user)
    await answer.location({'lat': 1, 'lng': 1}, session, user)

    assert trigger.triggered_times == 2


@pytest.mark.asyncio
async def test_can_combine_async_with_sync_parts():
    session = build_fake_session()
    user = build_fake_user()
    async_trigger = SimpleTrigger()
    sync_trigger = SimpleTrigger()

    @story.on('yo!')
    def one_story():
        @story.part()
        async def async_part(message):
            await asyncio.sleep(1)
            async_trigger.passed()

        @story.part()
        def sync_part(message):
            sync_trigger.passed()

    await answer.pure_text('yo!', session, user)
    assert async_trigger.is_triggered
    assert sync_trigger.is_triggered


@pytest.mark.asyncio
async def test_should_start_middlewares():
    story.use(mockdb.MockDB())
    http = story.use(mockhttp.MockHttpInterface())
    await story.start()
    http.start.assert_called_once_with()



import logging
import pytest

from . import messenger
from .. import mockdb, mockhttp
from ... import chat, story, utils
from ...middlewares import option

logger = logging.getLogger(__name__)


def teardown_function(function):
    logger.debug('tear down!')
    story.stories_library.clear()
    chat.interfaces = {}


@pytest.mark.asyncio
async def test_send_text_message():
    user = utils.build_fake_user()

    interface = story.use(messenger.FBInterface(token='qwerty'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await interface.send_text_message(
        recipient=user, text='hi!', options=None
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty',
        },
        json={
            'message': {
                'text': 'hi!',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_integration():
    user = utils.build_fake_user()

    story.use(messenger.FBInterface(token='qwerty'))
    story.use(mockdb.MockDB())
    mock_http = story.use(mockhttp.MockHttpInterface())

    await chat.say('hi there!', user=user)

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty',
        },
        json={
            'message': {
                'text': 'hi there!',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_options():
    user = utils.build_fake_user()

    story.use(messenger.FBInterface(token='qwerty'))
    story.use(mockdb.MockDB())
    mock_http = story.use(mockhttp.MockHttpInterface())

    await chat.ask(
        'Which color do you like?',
        options=[{
            'title': 'Red',
            'payload': 0xff0000,
        }, {
            'title': 'Green',
            'payload': 0x00ff00,
        }, {
            'title': 'Blue',
            'payload': 0x0000ff,
        }],
        user=user)

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty',
        },
        json={
            'message': {
                'text': 'Which color do you like?',
                'quick_replies': [
                    {
                        'content_type': 'text',
                        'title': 'Red',
                        'payload': 0xff0000,
                    },
                    {
                        'content_type': 'text',
                        'title': 'Green',
                        'payload': 0x00ff00,
                    },
                    {
                        'content_type': 'text',
                        'title': 'Blue',
                        'payload': 0x0000ff,
                    },
                ],
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_setup_webhook():
    fb_interface = story.use(messenger.FBInterface(webhook='/webhook'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    mock_http.webhook.assert_called_with(
        '/webhook',
        fb_interface.handle,
    )


# integration

@pytest.fixture
def build_fb_interface():
    async def builder():
        user = utils.build_fake_user()
        session = utils.build_fake_session()

        storage = story.use(mockdb.MockDB())
        fb = story.use(messenger.FBInterface(token='qwerty'))

        await storage.set_session(session)
        await storage.set_user(user)

        return fb

    return builder


@pytest.mark.asyncio
async def test_handler_raw_text(build_fb_interface):
    fb_interface = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on('hello, world!')
    def correct_story():
        @story.part()
        def store_result(message):
            correct_trigger.receive(message)

    @story.on('Goodbye, world!')
    def incorrect_story():
        @story.part()
        def store_result(message):
            incorrect_trigger.receive(message)

    await fb_interface.handle({
        "object": "page",
        "entry": [{
            "id": "PAGE_ID",
            "time": 1473204787206,
            "messaging": [
                {
                    "sender": {
                        "id": "USER_ID"
                    },
                    "recipient": {
                        "id": "PAGE_ID"
                    },
                    "timestamp": 1458692752478,
                    "message": {
                        "mid": "mid.1457764197618:41d102a3e1ae206a38",
                        "seq": 73,
                        "text": "hello, world!"
                    }
                }
            ]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'user': fb_interface.storage.user,
        'session': fb_interface.storage.session,
        'data': {
            'text': {
                'raw': 'hello, world!'
            }
        }
    }


@pytest.mark.asyncio
async def test_handler_selected_option(build_fb_interface):
    fb_interface = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on(receive=option.Match('GREEN'))
    def correct_story():
        @story.part()
        def store_result(message):
            correct_trigger.receive(message)

    @story.on(receive=option.Match('BLUE'))
    def incorrect_story():
        @story.part()
        def store_result(message):
            incorrect_trigger.receive(message)

    await fb_interface.handle({
        "object": "page",
        "entry": [{
            "id": "PAGE_ID",
            "time": 1473204787206,
            "messaging": [{
                "sender": {
                    "id": "USER_ID"
                },
                "recipient": {
                    "id": "PAGE_ID"
                },
                "timestamp": 1458692752478,
                "message": {
                    "mid": "mid.1457764197618:41d102a3e1ae206a38",
                    "seq": 73,
                    "text": "Green!",
                    "quick_reply": {
                        "payload": "GREEN"
                    }
                }
            }]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'user': fb_interface.storage.user,
        'session': fb_interface.storage.session,
        'data': {
            'option': 'GREEN',
            'text': {
                'raw': 'Green!'
            }
        }
    }



import logging
from . import processor
from .. import matchers

logger = logging.getLogger(__name__)


@matchers.matcher()
class WaitForReturn:
    type = 'WaitForReturn'

    def __init__(self):
        pass

    def validate(self, message):
        # return 'return' in message
        return True


class EndOfStory:
    type = 'EndOfStory'

    def __init__(self, data={}):
        self.data = data


def process_end_of_story(message, waiting_for):
    logger.debug('  got EndOfStory!')
    if message:
        message['data'] = {**message['data'], **waiting_for.data}
    return waiting_for


class CallableNodeWrapper:
    """
    helps start processing callable story
    """

    def __init__(self, ast_node, processor_instance):
        self.ast_node = ast_node
        self.processor_instance = processor_instance

    async def startpoint(self, *args, **kwargs):
        if 'session' not in kwargs:
            raise AttributeError('Got {} and {}. Should pass session as well'.format(args, kwargs))

        session = kwargs.pop('session')

        # we are going deeper so prepare one more item in stack
        logger.debug('  action: extend stack by +1')
        session['stack'].append(processor.build_empty_stack_item())
        res = await self.processor_instance.process_story(session,
                                                          # we don't have message yet
                                                          message=None,
                                                          compiled_story=self.ast_node,
                                                          idx=0,
                                                          story_args=args,
                                                          story_kwargs=kwargs)

        if isinstance(res, EndOfStory):
            return res.data
        return res


class CallableStoriesAPI:
    def __init__(self, library, parser_instance, processor_instance):
        self.library = library
        self.parser_instance = parser_instance
        self.processor_instance = processor_instance

    def callable(self):
        def fn(callable_story):
            compiled_story = self.parser_instance.compile(
                callable_story,
            )
            self.library.add_callable(compiled_story)
            return CallableNodeWrapper(
                compiled_story,
                self.processor_instance
            ).startpoint

        return fn



from ... import matchers


def is_location(message):
    return message.get('data', {}).get('location', False)


@matchers.matcher()
class Any:
    type = 'Location.Any'

    def __init__(self):
        pass

    def validate(self, message):
        return is_location(message)



import aiohttp
import logging
import pytest

import botstory.integrations.fb.messenger
from . import story
from . import chat
from .utils import answer, build_fake_session, build_fake_user, SimpleTrigger

logger = logging.getLogger(__name__)


def setup_function(function):
    logger.debug('setup')
    chat.interfaces = {}
    story.stories_library.clear()


def teardown_function(function):
    logger.debug('tear down!')
    chat.interfaces = {}
    story.stories_library.clear()


@pytest.fixture()
def mock_interface(mocker):
    mocker.patch.object(
        botstory.integrations.fb.messenger.FBInterface,
        'send_text_message',
        aiohttp.test_utils.make_mocked_coro('something'),
    )
    return chat.add_interface(botstory.integrations.fb.messenger.FBInterface())


@pytest.mark.asyncio
async def test_should_say(mock_interface):
    session = build_fake_session()
    user = build_fake_user()

    @story.on('hi there!')
    def one_story():
        @story.part()
        async def then(message):
            await chat.say('Nice to see you!', message['user'])

    await answer.pure_text('hi there!', session, user)

    mock_interface.send_text_message.assert_called_once_with(
        recipient=user,
        text='Nice to see you!',
    )


# TODO: move to middlewares/location/test_location.py
@pytest.mark.asyncio
@pytest.mark.skip
async def test_ask_location(mock_interface):
    session = build_fake_session()
    user = build_fake_user()

    @story.on('SOS!')
    def one_story():
        @story.part()
        async def then(message):
            await chat.ask_location('Hey, bro! Where is your rocket?', message['user'])

    await answer.pure_text('SOS!', session, user)

    mock_interface.send_text_message.assert_called_once_with(user, text='Hey, bro! Where is your rocket?')


@pytest.mark.asyncio
@pytest.mark.skip
async def test_get_location_as_result_of_asking_of_location(mock_interface):
    session = build_fake_session()
    user = build_fake_user()

    trigger = SimpleTrigger()

    @story.on('SOS!')
    def one_story():
        @story.part()
        def then(message):
            return chat.ask_location('Hey, bro! Where is your rocket?', message['user'])

        @story.part()
        def then(message):
            trigger.receive(message['data']['location'])

    await answer.pure_text('SOS!', session, user)
    await answer.location('somewhere', session, user)

    assert trigger.result() == 'somewhere'



from .. import story


async def location(loc, session=None, user=None):
    return await story.match_message({
        'data': {
            'location': loc,
        },
        'session': session,
        'user': user
    })


async def pure_text(text, session=None, user=None):
    return await story.match_message({
        'data': {
            'text': {'raw': text},
        },
        'session': session,
        'user': user
    })


async def option(payload, session=None, user=None):
    return await story.match_message({
        'data': {
            'option': payload,
        },
        'session': session,
        'user': user
    })



import logging
import json
import inspect

logger = logging.getLogger(__name__)


class Parser:
    def __init__(self):
        self.current_node = None
        self.middlewares = []

    def compile(self, one_story, middlewares=[]):
        topic = one_story.__name__
        self.middlewares = middlewares
        self.current_node = ASTNode(topic=topic)

        one_story()

        res = self.current_node
        self.current_node = None
        return res

    def go_deeper(self, one_story):
        if len(self.current_node.story_line) == 0 or \
                inspect.isfunction(self.current_node.story_line[-1]):
            self.current_node.story_line.append(StoryPartFork())

        parent_node = self.current_node
        child_story = self.compile(one_story, self.middlewares)
        parent_node.add_child(child_story)
        self.current_node = parent_node
        return child_story

    def part(self, story_part):
        for m in self.middlewares:
            if hasattr(m, 'process_part') and m.process_part(self, story_part):
                return True

        self.current_node.append(story_part)
        return True

    @property
    def topic(self):
        return self.current_node.topic


class ASTNode:
    def __init__(self, topic):
        self.compiled_story = None
        self.extensions = {}
        self.story_line = []
        self.story_names = set()
        self.topic = topic

    def add_child(self, child_story_line):
        """
        add child node to the last part of story
        :param child_story_line:
        :return:
        """
        assert isinstance(self.story_line[-1], StoryPartFork)
        self.story_line[-1].add_child(child_story_line)

    def append(self, story_part):
        part_name = story_part.__name__,
        if part_name in self.story_names:
            logger.warning('Already have story with name {}. Please use uniq name'.format(part_name))

        self.story_names.add(part_name)
        self.story_line.append(story_part)

    def to_json(self):
        return {
            'type': 'leaf',
            'topic': self.topic,
            'story_names': list(self.story_names),
        }

    def __str__(self):
        return json.dumps(self.to_json())


class StoryPartLeaf:
    def __init__(self, fn):
        self.fn = fn

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def __str__(self):
        return {
            'type': 'StoryPartLeaf',
            'name': self.__name__,
        }


class StoryPartFork:
    def __init__(self):
        self.children = []

    def __name__(self):
        return 'StoryPartFork'

    def add_child(self, child_story_line):
        self.children.append(child_story_line)

    def to_json(self):
        return {
            'type': self.__name__,
            'children': list(self.children),
        }

    def __str__(self):
        return json.dumps(self.to_json())



from unittest import mock
import aiohttp


def stub(name=None):
    """
    Creates a stub method. It accepts any arguments. Ideal to register to
    callbacks in tests.
    :param name: the constructed stub's name as used in repr
    :rtype: mock.MagicMock
    :return: Stub object.
    """
    return mock.MagicMock(spec=lambda *args, **kwargs: None, name=name)


class MockHttpInterface:
    post = aiohttp.test_utils.make_mocked_coro(return_value=True)
    webhook = stub('webhook')
    start = aiohttp.test_utils.make_mocked_coro(return_value=True)
    stop = aiohttp.test_utils.make_mocked_coro(return_value=True)
